#include "io.h"
extern "C" { // Tato část je vždy v čistém C - propojujeme kód v C a C++
  extern void Reset_Handler (void);
}

/// @file
/// @brief Zdrojový kód v C++
/* Základní IO použité v programu. Je to podobné jako v čistém C, jen je použita
 * jiná syntaxe, typická pro C++. Je to stejně potenciálně nebezpečný kus kódu,
 * tato syntaxe jen umožňuje ji lépe vyhledávat ve zdrojácích. Pokud takové přetypování
 * napíšeme, dáváme tím najevo, že víme přesně co děláme a byl by problém udělat to jinak.
 */
/// Pevný ukazatel na strukturu RCC_Type
static RCC_Type  * const RCC   = reinterpret_cast<RCC_Type  * const> RCC_BASE;
/// Pevný ukazatel na strukturu GPIO_Type
static GPIO_Type * const GPIOX = reinterpret_cast<GPIO_Type * const> LEDPORT_BASE;

/* I když to napíšeme jako třídu v C++, nic se nezmění.
 *  Po optimalizaci kód nijak nenaroste. Dá se dokonce říct,
 *  (alespoň u mé verze GCC to tak je), že děláme úplně stejnou
 *  věc a výsledná binárka je tedy úplně stejná jako v případě
 *  C-kódu. Protože jsou všechny metody jednoduché můžeme je nechat
 *  přímo v deklaraci třídy (deklarace je zároveň i definicí), takže
 *  budou implicitně inline - stejně jako v C.
 */
#include "bitindex.h"
/// GPIOX výstup jako C++ třída.
class GPIOCC {
  public:
    /// Konstruktor @param pin číslo pinu na bráně C
    GPIOCC (const unsigned pin) : bitpos (1U << pin) {
      // mělo by být nastavení pomocí |=, ale zde jen = nevadí, kratší o 2 instr.
      RCC  ->AHBENR = LEDCLK_GPIOEN;       // zavést hodiny
      GPIOX->MODER  = (1U << 2*pin);       // a nastavit jako výstup
      //TwoBitIndexable tbi (&GPIOX->MODER); // předchozí zápis jinak - ukázka, g++ to nepřebere.
      //tbi [pin] = 1;                       // MODER je indexované pole dvoubitových hodnot
    }
    /// Změna stavu pinu - metoda je konstantní, nemění data třídy.
    void toggle (void) const {
      GPIOX->ODR ^= bitpos;                // změnit výstup
    }
    /// Můžeme přidávat další metody..., pokud se v příkladu nepoužijí, gc-sections je odstraní.
    /// To je další podstatná výhoda C++, umožňuje nastavovat i složité chování periferie
    /// aniž by to pak uživatele obtěžovalo nebo mátlo a zapouzdřit je do třídy. Pro názornost
    /// je použit default parametr - další vlastnost C++, která v čistém C chybí.
    /// @param state Logická hodnota, na níž se má nastavit výstup.
    void set (const bool state = true) const {          // nepoužito
      if (state) GPIOX->BSRRL = bitpos;
      else       GPIOX->BSRRH = bitpos;
    }
  private:
    /// privátně jen konstantní pozice pinu
    const unsigned bitpos;
};

/// Zpoždění - poněkud špatně definované, ale stačí
/// @param tick - počet průchodů (délka průchodu je právě ne příliš dobře definována)
static inline void Delay (volatile unsigned int tick) {
  while (tick--);
}
// main() není potřeba, voláme přímo toto
/*NAKED*/ void Reset_Handler (void) {
  /* Instance GPIOCC je v tomto případě vytvořena na zásobníku z těchto důvodů :
   * 1. při vyšší optimalizaci se data třídy stejně nepoužijí,
   * 2. pokud by byla třída statická, bylo by nutné řešit volání statického konstruktoru,
   * což není triviální a zbytečně by to komplikovalo tento jednoduchý příklad.
   * (3. možností je vytvořit objekt na haldě, ale to je zcela mimo toto pojednání.)
   */
  GPIOCC led (LEDPIN);
  for (;;) {
    led.toggle ();
    Delay (DELAY_TIME);
  }
}
